//{{NO_DEPENDENCIES}}
// Microsoft Developer Studio generated include file.
// Used by gww.rc
//
#define GWW_ICON_ALT_TAB                101
#define GWW_ICON_TITLE_BAR              102
#define GWW_ICON_LOADING                105
#define SAR_ICON2                       105
#define GWW_ICON_SCORES                 106
#define SAR_ICON3                       106
#define GWW_ICON_JOYSTICK               107
#define GWW_ICON_SOUND                  108
#define GWW_ICON_CONFIG                 109
#define SAR_ICON4                       110
#define SAR_ICON                        111
#define SAR_ICON5                       112

// Next default values for new objects
// 
#ifdef APSTUDIO_INVOKED
#ifndef APSTUDIO_READONLY_SYMBOLS
#define _APS_NEXT_RESOURCE_VALUE        113
#define _APS_NEXT_COMMAND_VALUE         40001
#define _APS_NEXT_CONTROL_VALUE         1000
#define _APS_NEXT_SYMED_VALUE           101
#endif
#endif
