//Microsoft Developer Studio generated resource script.
//
#include "resource.h"

#define APSTUDIO_READONLY_SYMBOLS
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 2 resource.
//
#include "afxres.h"

/////////////////////////////////////////////////////////////////////////////
#undef APSTUDIO_READONLY_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
// English (U.S.) resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ENU)
#ifdef _WIN32
LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_US
#pragma code_page(1252)
#endif //_WIN32

#ifdef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// TEXTINCLUDE
//

1 TEXTINCLUDE DISCARDABLE 
BEGIN
    "resource.h\0"
END

2 TEXTINCLUDE DISCARDABLE 
BEGIN
    "#include ""afxres.h""\r\n"
    "\0"
END

3 TEXTINCLUDE DISCARDABLE 
BEGIN
    "\r\n"
    "\0"
END

#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// Icon
//

// Icon with lowest ID value placed first to ensure application icon
// remains consistent on all systems.
GWW_ICON_ALT_TAB        ICON    DISCARDABLE     "gwdata\\alt_tab.ico"
GWW_ICON_TITLE_BAR      ICON    DISCARDABLE     "gwdata\\title_bar.ico"
SAR_ICON2               ICON    DISCARDABLE     "icons\\SearchAndRescue2.ico"
SAR_ICON3               ICON    DISCARDABLE     "icons\\SearchAndRescue3.ico"
GWW_ICON_JOYSTICK       ICON    DISCARDABLE     "gwdata\\joystick.ico"
GWW_ICON_SOUND          ICON    DISCARDABLE     "gwdata\\sound.ico"
GWW_ICON_CONFIG         ICON    DISCARDABLE     "gwdata\\config.ico"
SAR_ICON4               ICON    DISCARDABLE     "icons\\SearchAndRescue4.ico"
SAR_ICON                ICON    DISCARDABLE     "icons\\SearchAndRescue.ico"
SAR_ICON5               ICON    DISCARDABLE     "icons\\SearchAndRescue5.ico"
#endif    // English (U.S.) resources
/////////////////////////////////////////////////////////////////////////////



#ifndef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 3 resource.
//


/////////////////////////////////////////////////////////////////////////////
#endif    // not APSTUDIO_INVOKED

