/*
	SAR Simulation Ending

	Switches from simulation back to menus, if a mission was being
	played then the mission results will be tabulated.
 */

#ifndef SARSIMEND_H
#define SARSIMEND_H

#include "sar.h"

extern void SARSimEnd(sar_core_struct *core_ptr);

#endif	/* SARSIMEND_H */
